"""this module used to implements the api for control the agent's actions include data collect and upload

"""

import atexit
import sys
import os
import time
import logging
import threading

from tingyun.config.settings import global_settings
from tingyun.packages import six
from tingyun.logistics.mapper import CONSTANCE_OUT_DATE_CONFIG, CONSTANCE_INVALID_DATA_TOKEN
from tingyun.armoury.sampler.memory_usage import memory_usage_sampler
from tingyun.armoury.sampler.cpu_usage import cpu_usage_sampler
from tingyun.battlefield.application import Application
from tingyun.packages.wrapt.decorators import synchronized

console = logging.getLogger(__name__)


class Dispatcher(object):
    """it is a agent core function center, include control function about the agent
    """
    _instance = None
    _instance_lock = threading.Lock()

    def __init__(self, config=None):
        """
        """
        console.info("Start python agent with version %s", global_settings().agent_version)
        self._last_harvest = 0.0
        # used to record the application connect status. avoid more connection thread to server .
        self._connect_status = []

        self._harvest_thread = threading.Thread(target=self._harvest_loop, name="TingYun-Harvest-Thread")
        self._harvest_thread.setDaemon(True)  # should take caution to the doc
        self._harvest_shutdown = threading.Event()

        self._config = config if config is not None else global_settings()
        self._lock = threading.Lock()
        self._process_shutdown = False
        self._applications = {}  # used to store the actived applications
        self.__max_tracker = 5000
        self.__data_sampler = []

        if self._config.enabled:
            atexit.register(self._atexit_shutdown)

            # Register our exit event to the uwsgi shutdown hooks.
            if 'uwsgi' in sys.modules:
                import uwsgi
                uwsgi_original_atexit_callback = getattr(uwsgi, 'atexit', None)

                def uwsgi_atexit_callback():
                    self._atexit_shutdown()
                    if uwsgi_original_atexit_callback:
                        uwsgi_original_atexit_callback()

                uwsgi.atexit = uwsgi_atexit_callback

    @staticmethod
    def singleton_instance():
        """
        :return: singleton instance
        """
        if Dispatcher._instance:
            return Dispatcher._instance

        instance = None
        with Dispatcher._instance_lock:
            if not Dispatcher._instance:
                instance = Dispatcher(global_settings())
                Dispatcher._instance = instance

                console.info('Creating instance of agent dispatcher')

        if instance:
            instance.register_data_sampler(memory_usage_sampler, ('memory.usage', ))
            instance.register_data_sampler(cpu_usage_sampler, ('cpu.usage', ))
            instance.active_dispatcher()

        return instance

    def global_settings(self):
        """
        :return:
        """
        return global_settings()

    def application_settings(self, app_name):
        """
        :param app_name:
        :return:
        """
        application = self._applications.get(app_name, None)
        if application:
            return application.application_config

    def _harvest_loop(self):
        """the real metric data harvest thread
        :rtype : None
        """
        console.info("start harvest thread with id %s", os.getpid())

        self.__next_harvest = time.time()
        last_harvest = time.time()
        try:
            while 1:
                # it's not a bug, just for deal the special situation
                if self._harvest_shutdown.isSet():  # python2.6 earlier api
                    self._do_harvest(last_harvest, time.time(), shutdown=True)
                    console.info("exit the controller first loop.")
                    return

                # we are not going to report in 1 min strictly, because the application or data maybe too large,
                # so extend the time appropriately according to the last report situation
                now = time.time()
                while self.__next_harvest <= now:
                    self.__next_harvest += 60.0

                self._harvest_shutdown.wait(self.__next_harvest - now)

                if self._harvest_shutdown.isSet():  # force do last harvest
                    self._do_harvest(last_harvest, time.time(), shutdown=True)
                    console.info("exit controller harvest loop at shutdown signal with last harvest.")
                    return

                self._do_harvest(last_harvest, time.time())
                last_harvest = time.time()
        except Exception as err:
            console.warning("error occurred %s", err)
            if self._process_shutdown:  # python interpreter exit
                console.critical('''Unexpected exception in main harvest loop when process being shutdown. This can occur
                                 in rare cases due to the main thread cleaning up and destroying objects while the
                                 background harvest thread is still running. If this message occurs rarely,
                                 it can be ignored. If the message occurs on a regular basis,
                                 then please report it to Ting Yun support for further investigation.''')
            else:                       # other error.
                console.critical('''Unexpected exception in main harvest loop. Please report this problem to Ting Yun
                                support for further investigation''')

    def _do_harvest(self, last_harvest, current_harvest, shutdown=False):
        """do the really harvest action
        :param shutdown: sign the agent status, shutdown or not
        :return:
        """
        self._last_harvest = time.time()

        for name, application in six.iteritems(self._applications):

            # ignore the disabled application
            if (not application.application_config) or (not application.application_config.enabled):
                return

            try:
                console.debug("Harvest data for application %s", name)
                ret = application.harvest(last_harvest, current_harvest, shutdown)
                if ret and (CONSTANCE_OUT_DATE_CONFIG == ret[0] or CONSTANCE_INVALID_DATA_TOKEN == ret[0]):
                    application.stop_connecting()
                    application.activate_session()
            except Exception as err:
                console.exception("Errors occurred when harvest application %s, %s", name, err)

        console.info("Spend %.2fs to harvest all applications.", time.time() - self._last_harvest)

    def active_dispatcher(self):
        """active a background controller thread if agent is enabled in settings
        """
        console.info('Start Agent controller main thread in process %s', os.getpid())
        if not self._config.enabled:
            console.info("agent is disabled,  agent controller not started.")
            return

        try:
            if self._harvest_thread.isAlive():
                console.info("agent controller was started, skip active it now.")
            else:
                console.info("Starting harvest thread now...")
                self._harvest_thread.start()

        except Exception as err:
            self._process_shutdown = True
            self.close_dispatcher()
            console.fatal("This exception indicates maybe internal agent python code error. please report to us"
                          " for further investigation. thank u.")
            console.fatal("Agent will stop work in this thread %s, %s", os.getpid(), err)

    def register_data_sampler(self, sampler, *args):
        """
        :param sampler:
        :return:
        """

        with self._lock:
            self.__data_sampler.append((sampler, args))

    @synchronized
    def active_application(self, app_name, link_app=""):
        """active the core application for collecting or report
        :param app_name: the application name. Note, current stage only one application supported
        :param link_app: the application linked application
        :return: None
        """
        # this should be local settings
        if not self._config.enabled:
            console.info("Agent was disabled in configure file.")
            return

        # the connection thread is working(always), so skip it
        if app_name in self._connect_status:
            return

        active_app = None
        # with self._lock:
        app = self._applications.get(app_name, None)
        try:
            if not app:
                app = Application(app_name, link_app)
                self._applications[app_name] = app
                self._connect_status.append(app_name)
                active_app = True
        except Exception as err:
            console.error("Init application failed, %s", err)

        for s in self.__data_sampler:
            app.register_data_sampler(s[0], s[1])

        if active_app:
            app.activate_session()

    def _atexit_shutdown(self):
        """define for python interpreter exit event
        :return:
        """

        self._process_shutdown = True
        self.close_dispatcher()

    def close_dispatcher(self, timeout=None):
        """shutdown the controller through the event signal
        """
        if timeout is None:
            timeout = self._config.shutdown_timeout

        if not self._harvest_shutdown.isSet():
            return

        # stop the connecting thread, if has.
        for name, application in six.iteritems(self._applications):
            application.stop_connecting()

        self._harvest_shutdown.set()
        self._harvest_thread.join(timeout)

        console.info('Tingyun agent is Shutdown...')

    def record_tracker(self, app_name, tracker_node):
        """
        """
        application = self._applications.get(app_name, None)
        if application is None:
            console.warning("Application %s not exist, if this is not expected, please contact us for help.", app_name)
            return

        # server disabled or agent not registered, skip to record the tracker data
        if not application.active:
            console.debug("Agent not registered to agent server, skip to record tracker data.")
            return

        if not application.application_config.enabled:
            console.debug("Agent disabled by server side, skip to record tracker data.")
            return

        application.record_tracker(tracker_node)


def dispatcher_instance():
    """
    :return: the singleton dispatcher object
    """
    dispatcher = Dispatcher.singleton_instance()
    return dispatcher




"""define a hook module for framework web2py. only support 0.3
"""

import logging
from tingyun.logistics.basic_wrapper import trace_out_function
from tingyun.armoury.trigger.wsgi_entrance import wsgi_app_wrapper_entrance
from tingyun.armoury.ammunition.webpy_tracker import trace_app_views
from tingyun.armoury.ammunition.function_tracker import wrap_function_trace

console = logging.getLogger(__name__)


def detect_wsgi_entrance(module):
    """
    :param module:
    :return:
    """
    trace_out_function(module, 'application.wsgifunc', wsgi_app_wrapper_entrance)


def detect_application(module):
    """
    :param module:
    :return:
    """
    # we caught the error through try/catch in views.
    trace_app_views(module, 'application._delegate')


def detect_app_template(module):
    """
    :param module:
    :return:
    """
    def template_name(instance, name):
        """
        :param instance: the instance of the Render
        :return: the template name
        """
        return name

    wrap_function_trace(module, "Render.__getattr__", name=template_name, group='Template.Render')
    wrap_function_trace(module, "Template.compile_template", name=template_name, group='Template.compile')




"""this module defined to get the memory usage information
"""

from tingyun.armoury.sampler.system_info import memory_used
from tingyun.logistics.attribution import TimeMetric


class MemoryUsage(object):
    """
    """
    def __init__(self, *args):
        """
        :return:
        """
        pass

    def start(self):
        """
        :return:
        """
        pass

    def stop(self):
        """
        :return:
        """
        pass

    def __call__(self, *args, **kwargs):
        """
        :param args:
        :param kwargs:
        :return:
        """
        metric_name = 'GENERAL/Memory/NULL/PhysicalUsed'
        memory = int(memory_used())

        yield [TimeMetric(name=metric_name, scope=metric_name, duration=memory, exclusive=memory), ]


memory_usage_sampler = MemoryUsage




"""Define this module for basic armory for flask

"""

import logging

from tingyun.armoury.trigger.wsgi_entrance import wsgi_application_wrapper
from tingyun.armoury.ammunition.function_tracker import wrap_function_trace
from tingyun.armoury.ammunition.flask_tracker import add_url_rule_wrapper, handle_exception_wrapper
from tingyun.armoury.ammunition.flask_tracker import wrap_register_error, wrap_func_with_request
from tingyun.logistics.basic_wrapper import wrap_function_wrapper

console = logging.getLogger(__name__)


def detect_wsgi_entrance(module):
    """
    :param module:
    :return:
    """
    version = 'xx'
    try:
        import flask
        version = getattr(flask, "__version__", 'xx')
    except Exception as _:
        pass

    wsgi_application_wrapper(module.Flask, '__call__', ('flask', version))


def detect_app_entrance(module):
    """
    :param module:
    :return:
    """
    # trace the views function call metric
    wrap_function_wrapper(module.Flask, 'add_url_rule', add_url_rule_wrapper)
    wrap_function_wrapper(module.Flask, 'handle_exception', handle_exception_wrapper)

    # new feature form flask 0.7
    if hasattr(module.Flask, 'handle_user_exception'):
        wrap_function_wrapper(module.Flask, 'handle_user_exception', handle_exception_wrapper)

    # new feature from flask 0.7
    # todo: some issue fix
    if hasattr(module.Flask, 'full_dispatch_request'):
        wrap_function_trace(module.Flask, 'full_dispatch_request')

    # new feature from flask 0.7. this will occurred when errorhandler or register_error_handler is used by application
    if hasattr(module.Flask, '_register_error_handler'):
        wrap_function_wrapper(module.Flask, '_register_error_handler', wrap_register_error)

    # deal for before_first_request, new feature from 0.9
    if hasattr(module.Flask, 'before_first_request'):
        wrap_function_wrapper(module.Flask, 'before_first_request', wrap_func_with_request)

    # deal for before_request
    if hasattr(module.Flask, 'before_request'):
        wrap_function_wrapper(module.Flask, 'before_request', wrap_func_with_request)

    # deal for after_request
    if hasattr(module.Flask, 'after_request'):
        wrap_function_wrapper(module.Flask, 'after_request', wrap_func_with_request)

    # deal for teardown_request
    if hasattr(module.Flask, 'teardown_request'):
        wrap_function_wrapper(module.Flask, 'teardown_request', wrap_func_with_request)


def detect_app_blueprint_entrance(module):
    # BluePrint
    # deal for blueprint. new feature form v0.7
    if hasattr(module.Blueprint, 'before_request'):
        console.info("before blure print request.")
        wrap_function_wrapper(module.Blueprint, 'before_request', wrap_func_with_request)

    if hasattr(module.Blueprint, 'before_app_request'):
        wrap_function_wrapper(module.Blueprint, 'before_app_request', wrap_func_with_request)

    if hasattr(module.Blueprint, 'before_app_first_request'):
        wrap_function_wrapper(module.Blueprint, 'before_app_first_request', wrap_func_with_request)

    if hasattr(module.Blueprint, 'after_request'):
        wrap_function_wrapper(module.Blueprint, 'after_request', wrap_func_with_request)

    if hasattr(module.Blueprint, 'after_app_request'):
        wrap_function_wrapper(module.Blueprint, 'after_app_request', wrap_func_with_request)

    if hasattr(module.Blueprint, 'teardown_request'):
        wrap_function_wrapper(module.Blueprint, 'teardown_request', wrap_func_with_request)

    if hasattr(module.Blueprint, 'teardown_app_request'):
        wrap_function_wrapper(module.Blueprint, 'teardown_app_request', wrap_func_with_request)


def detect_templates(module):
    """
    :param module:
    :return:
    """
    wrap_function_trace(module, '_render')



"""This module implements functions for querying properties of the operating
system or for the specific process the code is running in.

"""

import os
import sys
import multiprocessing
import re

try:
    import resource
except ImportError:
    pass


def cpu_count():
    """Returns the number of CPUs in the system.

    """
    # The multiprocessing module support Windows, BSD systems  MacOS X and systems which support the POSIX API
    try:
        return multiprocessing.cpu_count()
    except NotImplementedError:
        pass

    # For Jython, we need to query the Java runtime environment.
    try:
        from java.lang import Runtime
        runtime = Runtime.getRuntime()
        res = runtime.availableProcessors()
        if res > 0:
            return res
    except ImportError:
        pass

    # Solaris system
    try:
        pseudoDevices = os.listdir('/devices/pseudo/')
        expr = re.compile('^cpuid@[0-9]+$')

        res = 0
        for pd in pseudoDevices:
            if expr.match(pd) != None:
                res += 1

        if res > 0:
            return res
    except OSError:
        pass

    return 1


def memory_total():
    """Returns the total physical memory available in the system.

    """
    # For Linux we can determine it from the proc filesystem.
    if sys.platform == 'linux2':
        try:
            parser = re.compile(r'^(?P<key>\S*):\s*(?P<value>\d*)\s*kB')

            with open('/proc/meminfo') as fp:
                try:
                    for line in fp.readlines():
                        match = parser.match(line)
                        if not match:
                            continue
                        key, value = match.groups(['key', 'value'])
                        if key == 'MemTotal':
                            memory_bytes = float(value) * 1024
                            return memory_bytes / (1024*1024)
                except Exception as _:
                    pass
        except IOError:
            pass

    # other platforms
    try:
        import psutil
        return psutil.virtual_memory().total
    except (ImportError, AttributeError):
        pass

    return 0


def memory_used():
    """Returns the memory used in MBs. Calculated differently depending
    on the platform and designed for informational purposes only.

    """
    if sys.platform == 'linux2':
        pid = os.getpid()
        statm = '/proc/%d/statm' % pid

        with open(statm, 'r') as fp:
            try:
                rss_pages = float(fp.read().split()[1])
                memory_bytes = rss_pages * resource.getpagesize()

                return memory_bytes / (1024*1024)
            except Exception as _:
                pass
    return 0


def cpu_info():
    """calculate the cpu information
    :return:(cpu_vendor, cpu_model, cpu_mhz)
    """
    cpu_file = "/proc/cpuinfo"
    ret = ["unknown", "unknown", "unknown"]

    try:
        infos = {}
        with open(cpu_file, 'r') as f:
            for line in f:
                kv = line.split(":")
                if len(kv) != 2:
                    continue

                key = kv[0].strip()
                value = kv[1].strip()
                if key in infos:
                    break

                infos[key] = value

        ret[0] = infos['vendor_id']
        ret[1] = infos['model name']
        ret[2] = infos['cpu MHz']
    except Exception as _:
        pass

    return ret



""" define some detector for tracer entrance
"""

import time
import logging

from tingyun.armoury.ammunition.tornado_4.utils import is_websocket, generate_tracer
from tingyun.logistics.basic_wrapper import wrap_function_wrapper
from tingyun.armoury.ammunition.tracker import current_tracker
from tingyun.logistics.object_name import callable_name

console = logging.getLogger(__name__)


def trace_server_request_init_(wrapped, request, args, kwargs):
    """agent tracer entrance in tornado, will be called  after header but body. Three situations will be called.
        1. a common web.Application instance.
        2.instance passed into httpserver.HTTPServer is a not tornado application.
        3. In wsgi module. some WSGI applications.

    :param request: a `request` Instance of current user request.
    """
    result = wrapped(*args, **kwargs)

    tracker = None if is_websocket(request) else generate_tracer(request)
    if tracker:
        tracker.set_tracker_name(callable_name(wrapped), priority=1)
        tracker.start_time = getattr(request, "_start_time", int(time.time()))

    request._self_tracker = tracker
    return result


def trace_server_request_finish(wrapped, instance, args, kwargs):
    """in WSGI compatible in tornado. this should be export
    :param wrapped:
    :param instance:
    :param args:
    :param kwargs:
    :return:
    """
    tracker = current_tracker()
    if not tracker:
        console.debug("current tracker is gone, return response now.")
        return wrapped(*args, **kwargs)

    try:
        tracker.finish_work(None, None, None)
    finally:
        return wrapped(*args, **kwargs)


def detect_tracker_entrance(module):
    """
    :param module:
    :return:
    """
    wrap_function_wrapper(module, "HTTPServerRequest.__init__", trace_server_request_init_)
    # wrap_function_wrapper(module, "HTTPServerRequest.finish", trace_server_request_finish)
